from sequence.topology.router_net_topo import RouterNetTopo
# from sequence.topology.node import Node, BSMNode, QuantumRouter
from sequence.topology.topology import Topology as Topo
import random
import math
from networkx import Graph, all_shortest_paths, shortest_simple_paths, exception, shortest_path_length
from graph_builder import is_high
import json
import os
def dict_to_topo(dictionary) -> RouterNetTopo:
    with open("stuff.json", "w") as f:
        json.dump(dictionary, f)
    nw=RouterNetTopo("stuff.json")
    os.remove("stuff.json")
    return nw
        
def set_parameters(topology: RouterNetTopo):
    # set memory parameters
    MEMO_FREQ = 2e3
    MEMO_EXPIRE = 0
    MEMO_EFFICIENCY = 1
    MEMO_FIDELITY = 0.975
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
        memory_array = node.get_components_by_type("MemoryArray")[0] 
        memory_array.update_memory_params("frequency", MEMO_FREQ)
        memory_array.update_memory_params("coherence_time", MEMO_EXPIRE)
        memory_array.update_memory_params("efficiency", MEMO_EFFICIENCY)
        memory_array.update_memory_params("raw_fidelity", MEMO_FIDELITY)

    # set detector parameters
    DETECTOR_EFFICIENCY = 0.9
    DETECTOR_COUNT_RATE = 5e7
    DETECTOR_RESOLUTION = 100
    for node in topology.get_nodes_by_type(RouterNetTopo.BSM_NODE):
        bsm = node.get_components_by_type("SingleAtomBSM")[0]
        bsm.update_detectors_params("efficiency", DETECTOR_EFFICIENCY)
        bsm.update_detectors_params("count_rate", DETECTOR_COUNT_RATE)
        bsm.update_detectors_params("time_resolution", DETECTOR_RESOLUTION)
    # set entanglement swapping parameters
    SWAP_SUCC_PROB = 0.90
    SWAP_DEGRADATION = 0.99
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
        node.network_manager.protocol_stack[1].set_swapping_success_rate(SWAP_SUCC_PROB)
        node.network_manager.protocol_stack[1].set_swapping_degradation(SWAP_DEGRADATION)
        
    # set quantum channel parameters
    ATTENUATION = 1e-5
    QC_FREQ = 1e11
    for qc in topology.get_qchannels():
        qc.attenuation = ATTENUATION
        qc.frequency = QC_FREQ


def set_efficiency_xi(topology: RouterNetTopo, xi: float):
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
        MEMO_EFFICIENCY = 0.999 if random.random() < xi else 0.8
        memory_array = node.get_components_by_type("MemoryArray")[0]
        memory_array.update_memory_params("efficiency", MEMO_EFFICIENCY)
def set_efficiency_alpha(topology: RouterNetTopo, alpha: float):
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
        MEMO_EFFICIENCY = math.log(random.uniform(math.e**(0.8*alpha), math.e**(0.999*alpha)))/alpha
        memory_array = node.get_components_by_type("MemoryArray")[0]
        memory_array.update_memory_params("efficiency", MEMO_EFFICIENCY)


def clear_forwarding_tables(topology: RouterNetTopo):
    for src in topology.nodes[topology.QUANTUM_ROUTER]:
        routing_protocol = src.network_manager.protocol_stack[0]
        routing_protocol.forwarding_table = {}


def gen_tables_shortest_path(topology: RouterNetTopo):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values()) 
    for src in topology.nodes[topology.QUANTUM_ROUTER]:
        for dst_name in graph.nodes:
            if src.name == dst_name:
                continue
            try:
                if dst_name > src.name:
                    paths = all_shortest_paths(graph, src.name, dst_name, weight=None)
                else:
                    paths = list(map(lambda l: l[::-1], all_shortest_paths(graph, dst_name, src.name, weight=None)))
                for path in paths:
                    resulting_fidelity = 0.975
                    for node in path[1:-1]:
                        resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                    if resulting_fidelity > 0.53:
                        next_hop = path[1]
                        # routing protocol locates at the bottom of the stack
                        routing_protocol = src.network_manager.protocol_stack[0]  # guarantee that [0] is the routing protocol?
                        routing_protocol.add_forwarding_rule(dst_name, next_hop)
                        break
            except exception.NetworkXNoPath:
                pass

def gen_tables_efficiency_cost(topology: RouterNetTopo, e_max=0.999, e_min=0.8):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values()) 
    for src in topology.nodes[topology.QUANTUM_ROUTER]:
        for dst_name in graph.nodes:
            if src.name == dst_name:
                continue
            try:
                def cost(x, y, edge_dict):
                        math.e**(10*(graph.nodes[y]["efficiency"] - e_min)/(e_max-e_min))
                paths=None
                if dst_name > src.name:
                    paths = all_shortest_paths(graph, src.name, dst_name, weight=cost)
                else:
                    paths = list(map(lambda l: l[::-1], all_shortest_paths(graph, dst_name, src.name, weight=cost)))
                for path in paths:
                    resulting_fidelity = 0.975
                    for node in path[1:-1]:
                        resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                    if resulting_fidelity > 0.53:
                        next_hop = path[1]
                        # routing protocol locates at the bottom of the stack
                        routing_protocol = src.network_manager.protocol_stack[0]  # guarantee that [0] is the routing protocol?
                        routing_protocol.add_forwarding_rule(dst_name, next_hop)
                        break
            except exception.NetworkXNoPath:
                pass


def gen_tables_kshortest_path(topology: RouterNetTopo, source, destination, k = 10):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values())
    if source==destination:
        return None
    try:
        paths=None
        if destination > source:
            paths=shortest_simple_paths(graph, source, destination, weight = None)
        else:
            paths=map(lambda l: l[::-1], shortest_simple_paths(graph, destination, source, weight = None))
        final_path = None
        min_fidelity = 10
        for (i, p) in enumerate(paths):
            if i>=k:
                break
            resulting_fidelity = 0.975
            for node in p[1:-1]:
                resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
            if resulting_fidelity < min_fidelity  and resulting_fidelity > 0.53:
                min_fidelity = resulting_fidelity
                final_path = p
        if final_path == None:
            raise exception.NetworkXNoPath
        for (i, node_name) in enumerate(final_path[:-1]):
            for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
                if node.name == node_name:
                    node.network_manager.protocol_stack[0].update_forwarding_rule(destination, final_path[i+1])
    except exception.NetworkXNoPath:
        pass

def gen_tables_kxshortest_path(topology: RouterNetTopo, source, destination, k = 10, x=1):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values())
    if source==destination:
        return None
    try:
        paths=None
        min_length=None
        if destination > source:
            paths=shortest_simple_paths(graph, source, destination, weight = None)
            min_length=shortest_path_length(graph, source, destination, weight = None)
        else:
            paths=map(lambda l: l[::-1], shortest_simple_paths(graph, destination, source, weight = None))
            min_length=shortest_path_length(graph, destination, source, weight = None)
        final_path = None
        min_fidelity = 10
        for (i, p) in enumerate(paths):
            if i>=k or abs(len(p) - (min_length+1)) > x:
                break
            resulting_fidelity = 0.975
            for node in p[1:-1]:
                resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
            if resulting_fidelity < min_fidelity  and resulting_fidelity > 0.53:
                min_fidelity = resulting_fidelity
                final_path = p
        if final_path == None:
            raise exception.NetworkXNoPath
        for (i, node_name) in enumerate(final_path[:-1]):
            for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
                if node.name == node_name:
                    node.network_manager.protocol_stack[0].update_forwarding_rule(destination, final_path[i+1])
    except exception.NetworkXNoPath:
        pass

def gen_tables_kshortest_path_qos(topology: RouterNetTopo, source, destination, k = 10, is_high=is_high):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values())
    if source==destination:
        return None
    try:
        paths=None
        if destination > source:
            paths=shortest_simple_paths(graph, source, destination, weight = None)
        else:
            paths=map(lambda l: l[::-1], shortest_simple_paths(graph, destination, source, weight = None))
        final_path=None
        if is_high(destination):
            min_fidelity = 10
            for (i, p) in enumerate(paths):
                if i>=k:
                    break
                resulting_fidelity = 0.975
                for node in p[1:-1]:
                    resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                if resulting_fidelity < min_fidelity  and resulting_fidelity > 0.53:
                    min_fidelity = resulting_fidelity
                    final_path = p
        else:
            max_fidelity = -1
            for (i, p) in enumerate(paths):
                if i>=k:
                    break
                resulting_fidelity = 0.975
                for node in p[1:-1]:
                    resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                if resulting_fidelity > max_fidelity  and resulting_fidelity > 0.53:
                    max_fidelity = resulting_fidelity
                    final_path = p
                
        if final_path == None:
            raise exception.NetworkXNoPath
        for (i, node_name) in enumerate(final_path[:-1]):
            for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
                if node.name == node_name:
                    node.network_manager.protocol_stack[0].update_forwarding_rule(destination, final_path[i+1])
    except exception.NetworkXNoPath:
        pass

def gen_tables_kxshortest_path_qos(topology: RouterNetTopo, source, destination, k = 10, x=1, is_high=is_high):
    graph = Graph()
    for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):     
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        graph.add_node(node.name, fidelity=fidelity, efficiency=efficiency)
    costs = {}
    for qc in topology.qchannels:
        router, bsm = qc.sender.name, qc.receiver
        if bsm not in costs:
            costs[bsm] = [router, qc.distance]
        else:
            costs[bsm] = [router] + costs[bsm]
            costs[bsm][-1] += qc.distance
    graph.add_weighted_edges_from(costs.values())
    if source==destination:
        return None
    try:
        paths=None
        min_length=None
        if destination > source:
            paths=shortest_simple_paths(graph, source, destination, weight = None)
            min_length=shortest_path_length(graph, source, destination, weight = None)
        else:
            paths=map(lambda l: l[::-1], shortest_simple_paths(graph, destination, source, weight = None))
            min_length=shortest_path_length(graph, destination, source, weight = None)
        final_path=None
        if is_high(destination):
            min_fidelity = 10
            for (i, p) in enumerate(paths):
                if i>=k or abs(len(p) - (min_length+1)) > x:
                    break
                resulting_fidelity = 0.975
                for node in p[1:-1]:
                    resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                if resulting_fidelity < min_fidelity  and resulting_fidelity > 0.53:
                    min_fidelity = resulting_fidelity
                    final_path = p
        else:
            max_fidelity = -1
            for (i, p) in enumerate(paths):
                if i>=k or abs(len(p) - (min_length+1)) > x:
                    break
                resulting_fidelity = 0.975
                for node in p[1:-1]:
                    resulting_fidelity = (resulting_fidelity-0.25)* ((4*graph.nodes[node]["efficiency"]**2 - 1)/3) * ((4*graph.nodes[node]["fidelity"] - 1)/3) + 0.25
                if resulting_fidelity > max_fidelity  and resulting_fidelity > 0.53:
                    max_fidelity = resulting_fidelity
                    final_path = p
                
        if final_path == None:
            raise exception.NetworkXNoPath
        for (i, node_name) in enumerate(final_path[:-1]):
            for node in topology.get_nodes_by_type(RouterNetTopo.QUANTUM_ROUTER):
                if node.name == node_name:
                    node.network_manager.protocol_stack[0].update_forwarding_rule(destination, final_path[i+1])
    except exception.NetworkXNoPath:
        pass



from graph_builder import *
from network_generator import *
from metrics import *

higher_better=0
lower_better=0
for i in range(1,101):
    nw1=dict_to_topo(regular_gen(6, frac=0.5))
    set_parameters(nw1)
    set_efficiency_xi(nw1, 0.9)
    mean_hp, stddev_hp, mean_lp, stddev_lp = evaluate_kx0shortest_path_qos(nw1)
    if mean_hp > mean_lp:
        higher_better+=1
    else:
        lower_better+=1

print(higher_better, lower_better)



import networkx as nx
import random
def regular_gen(n: int, frac=0.3) -> dict:
    '''
    Returns a network topology following a regular graph structure in the form of a dict.
    
    A "regular" graph is a graph with n sources, n destinations,
    and n**2 repeaters. Connected in such a fasion that there are n layers
    of n repeaters placed between the source and the destination.

    For eg. n = 2, a regular graph looks like this.

    s1 -- n11 -- n12 -- d1
          |      |
    s2 -- n21 -- n22 -- d2

    For n = 3,

       s1-n11----n13----n13-d1   
          | \\    | \\    | \\   
       s3-|--n31-|--n32-|--n33-d3
          | /    | /    | /   
       s2-n23----n23----n23-d2   

    '''

    # Dictionary structure referenced from: https://sequence-rtd-tutorial.readthedocs.io/en/latest/tutorial/chapter5/network_manager.html#step-1-create-the-network-configuration-file
    seed = 0
    graph = {
        "is_parallel": False,
        "stop_time": 2000000000000,
        "nodes": [],
        "qconnections": [],
        "cconnections": []
    }

    high_quality=random.sample(range(1, n+1), round(n*frac))
    
    # Generating all nodes.
    for i in range(1, n+1):
        src_dict = {
            "name": ("s"+str(i)),
            "type": "QuantumRouter",
            "seed": seed,
            "memo_size": 50
        }
        graph["nodes"].append(src_dict)
        seed+=1
        
        dest_dict = {
            "name": (("hd" if i in high_quality else "d") +str(i)),
            "type": "QuantumRouter",
            "seed": seed,
            "memo_size": 50
        }
        graph["nodes"].append(dest_dict)
        seed+=1
        
        for j in range(1, n+1):
            node_dict = {
                "name": ("n"+str(i)+"_"+str(j)),
                "type": "QuantumRouter",
                "seed": seed,
                "memo_size": 50
            }
            graph["nodes"].append(node_dict)
            seed+=1

    # Generating all edges.
    for i in range(1,n+1):
        q_edge_src = {
            "node1": "s"+str(i),
            "node2": "n"+str(i)+"_1",
            "attenuation": 0.0002,
            "distance": 500,
            "type": "meet_in_the_middle"
        }
        
        c_edge_src = {
            "node1": "s"+str(i),
            "node2": "n"+str(i)+"_1",
            "delay": 500000000
        }
        graph["qconnections"].append(q_edge_src)
        graph["cconnections"].append(c_edge_src)
        q_edge_dest = {
            "node1": ("hd" if i in high_quality else "d")+str(i),
            "node2": "n"+str(i)+"_"+str(n),
            "attenuation": 0.0002,
            "distance": 500,
            "type": "meet_in_the_middle"
        }
        c_edge_dest = {
            "node1": ("hd" if i in high_quality else "d")+str(i),
            "node2": "n"+str(i)+"_"+str(n),
            "delay": 500000000
        }
        graph["qconnections"].append(q_edge_dest)
        graph["cconnections"].append(c_edge_dest)
        
        for j in range(1, n+1):
            # Intralayer connection
            if n!=2 or i!=2:
                q_edge_intra = {
                    "node1": "n"+str(i)+"_"+str(j),
                    "node2": "n"+str(1 if i==n else i+1)+"_"+str(j),
                    "attenuation": 0.0002,
                    "distance": 500,
                    "type": "meet_in_the_middle"
                }
                c_edge_intra = {
                    "node1": "n"+str(i)+"_"+str(j),
                    "node2": "n"+str(1 if i==n else i+1)+"_"+str(j),
                    "delay": 500000000
                }
                graph["qconnections"].append(q_edge_intra)
                graph["cconnections"].append(c_edge_intra)
            # Interlayer connection
            if j != n:
                q_edge_inter = {
                    "node1": "n"+str(i)+"_"+str(j),
                    "node2": "n"+str(1)+"_"+str(j+1),
                    "attenuation": 0.0002,
                    "distance": 500,
                    "type": "meet_in_the_middle"
                }
                c_edge_inter = {
                    "node1": "n"+str(i)+"_"+str(j),
                    "node2": "n"+str(1)+"_"+str(j+1),
                    "delay": 500000000
                }
                graph["qconnections"].append(q_edge_inter)
                graph["cconnections"].append(c_edge_inter)
                
    return graph    
def waxman_gen(n: int, alpha = 0.85, beta=0.275, frac=0.3) -> dict:
    G = nx.waxman_graph(n**2, alpha = alpha, beta = beta)
    seed = 0
    graph = {
        "is_parallel": False,
        "stop_time": 2000000000000,
        "nodes": [],
        "qconnections": [],
        "cconnections": []
    }
    high_quality=random.sample(range(1, n+1), round(n*frac))
    for i in range(1, n**2+1):
        node_dict =  {
                "name": ("n"+str(i)),
                "type": "QuantumRouter",
                "seed": seed,
                "memo_size": 50
        }
        graph["nodes"].append(node_dict)
        seed +=1
        

    for k in range(1, n+1):
        src_dict =  {
                "name": ("s"+str(k)),
                "type": "QuantumRouter",
                "seed": seed,
                "memo_size": 50
        }
        seed+=1
        dest_dict =  {
                "name": (("hd" if k in high_quality else "d")+str(k)),
                "type": "QuantumRouter",
                "seed": seed,
                "memo_size": 50
        }
        
        seed +=1
        graph["nodes"].append(src_dict)
        graph["nodes"].append(dest_dict)
        
    for (u, v, _) in G.edges.data():
        pos1 = G.nodes[u]["pos"]
        pos2 = G.nodes[v]["pos"]
        q_edge_dict = {
            "node1": "n"+str(u+1),
            "node2": "n"+str(v+1),
            "attenuation": 0.0002,
            "distance": round(1000*((pos2[1]-pos1[1])**2 +  (pos2[0]-pos1[0])**2)**0.5 ) ,
            "type": "meet_in_the_middle"
        }
        c_edge_dict = {
            "node1": "n"+str(u+1),
            "node2": "n"+str(v+1),
            "delay": 500000000
        }
        graph["qconnections"].append(q_edge_dict)
        graph["cconnections"].append(c_edge_dict)

    src_nodes = random.sample(range(1, n**2+1), n)
    dest_nodes = random.sample(range(1, n**2+1), n)

    for i in range(1, n+1):
        q_edge_dict_src = {
            "node1": "s"+str(i),
            "node2": "n"+str(src_nodes[i-1]),
            "attenuation": 0.0002,
            "distance": 500,
            "type": "meet_in_the_middle"
        }
        c_edge_dict_src = {
            "node1": "s"+str(i),
            "node2": "n"+str(src_nodes[i-1]),
            "delay": 500000000
        }
        graph["qconnections"].append(q_edge_dict_src)
        graph["cconnections"].append(c_edge_dict_src)

        q_edge_dict_dest = {
            "node1": ("hd" if i in high_quality else "d")+str(i),
            "node2": "n"+str(dest_nodes[i-1]),
            "attenuation": 0.0002,
            "distance": 500,
            "type": "meet_in_the_middle"
        }
        c_edge_dict_dest = {
            "node1": ("hd" if i in high_quality else "d")+str(i),
            "node2": "n"+str(dest_nodes[i-1]),
            "delay": 500000000
        }
        graph["qconnections"].append(q_edge_dict_dest)
        graph["cconnections"].append(c_edge_dict_dest)
    return graph

def is_high(string: str) -> bool:
    if len(string) > 0:
        return string[0]=='h'
    else:
        return False



from sequence.topology.router_net_topo import RouterNetTopo
from graph_builder import is_high
import random
from network_generator import *

def get_node_path(topology: RouterNetTopo, source, destination):
    src_node = None
    dst_node = None
    for node in topology.nodes[topology.QUANTUM_ROUTER]:
        if source == node.name:
            src_node = node
        elif destination == node.name:
            dst_node = node

    if src_node is None or dst_node is None:
        return None

    path = []
    current_node = src_node
    while current_node.name != destination:
        path.append(current_node)
        table = current_node.network_manager.protocol_stack[0].forwarding_table
        if destination not in table.keys():
            return None
        else:
            next_hop = table[destination]
            for node in topology.nodes[topology.QUANTUM_ROUTER]:
                if next_hop == node.name:
                    current_node = node
                    break
                path.append(current_node)
    return path

def calculate_fidelity(topology: RouterNetTopo, source, destination):
    path = get_node_path(topology, source, destination)
    if path is None:
        return 0
    resulting_fidelity=0.975
    for node in path[1:-1]:
        fidelity= node.get_components_by_type("MemoryArray")[0][0].raw_fidelity
        efficiency= node.get_components_by_type("MemoryArray")[0][0].efficiency
        resulting_fidelity = (resulting_fidelity-0.25)*((4*efficiency**2-1)/3)*((4*fidelity-1)/3) + 0.25
    return resulting_fidelity
def print_forwarding_tables(topology: RouterNetTopo):
    for node in topology.nodes[topology.QUANTUM_ROUTER]:
        print(node.name, node.network_manager.protocol_stack[0].forwarding_table)
        
def get_source_nodes(topology: RouterNetTopo):
    nodes=[]
    for node in topology.nodes[topology.QUANTUM_ROUTER]:
        if 's' in node.name:
            nodes.append(node)
    return nodes

def get_dest_nodes(topology: RouterNetTopo):
    nodes=[]
    for node in topology.nodes[topology.QUANTUM_ROUTER]:
        if 'd' in node.name:
            nodes.append(node)
    return nodes

def find_mean(num_list):
    n = len(num_list)
    if n == 0:
        return -1
    sum = 0
    for i in num_list:
        sum+=i
    return sum/n
def find_stddev(num_list):
    n = len(num_list)
    if n == 0:
        return -1
    mean = find_mean(num_list)
    sum = 0
    for i in num_list:
        sum+= (i-mean)**2
    return (sum/n)**0.5

def evaluate_shortest_path(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    gen_tables_shortest_path(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))


def evaluate_efficiency_cost(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    gen_tables_efficiency_cost(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))

def evaluate_kshortest_path(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kshortest_path(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kshortest_path(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))

def evaluate_kxshortest_path(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kxshortest_path(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kxshortest_path(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))

def evaluate_kx0shortest_path(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kxshortest_path(topology, source.name, destination.name, x=0)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kxshortest_path(topology, source.name, destination.name, x=0)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))


def evaluate_kshortest_path_qos(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kshortest_path_qos(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kshortest_path_qos(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))

def evaluate_kxshortest_path_qos(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kxshortest_path_qos(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kxshortest_path_qos(topology, source.name, destination.name)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))


def evaluate_kx0shortest_path_qos(topology: RouterNetTopo, is_high=is_high, get_source_nodes=get_source_nodes, get_dest_nodes=get_dest_nodes):
    clear_forwarding_tables(topology)
    source_nodes = get_source_nodes(topology)
    dest_nodes = get_dest_nodes(topology)
    hp_nodes = list(filter(lambda x: is_high(x.name), dest_nodes))
    lp_nodes = list(filter(lambda x: not is_high(x.name), dest_nodes))
    hp_fidelities = []
    lp_fidelities = []
    for source in source_nodes:
        for destination in hp_nodes:
            gen_tables_kxshortest_path_qos(topology, source.name, destination.name, x=0)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            hp_fidelities.append(fidelity)
        for destination in lp_nodes:
            gen_tables_kxshortest_path_qos(topology, source.name, destination.name, x=0)
            fidelity = calculate_fidelity(topology, source.name, destination.name)
            lp_fidelities.append(fidelity)


    return (find_mean(hp_fidelities), find_stddev(hp_fidelities), find_mean(lp_fidelities), find_stddev(lp_fidelities))
